/*
Telemetrie für die Amateurfunkanlage von DL1SUZ auf Basis von Modbus/RTU
Version 0.4  Stand 2023-12-16-dl1suz

***************************************************************************************************************
 *
 *  This program is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License
 *
 *                            http://creativecommons.org/licenses/by-nc-sa/3.0/
 *
 *                        http://creativecommons.org/licenses/by-nc-sa/3.0/legalcode
 *
 *
 ***************************************************************************************************************

V0.40: Frequenzumschaltung für 23cm ergänzt
V0.30: Max_PWR für Skalierung des Bargraphes eingeführt

Vereinbarungen zum Modbus/RTU-Protokoll
Die Modbus-Kommunikation wird komplett über die die Holding-Register abgewickelt
40001:           DO[0] - DO[15]  - 16 Einzelbits für Kommunikation vom Master zum Slave 
                 reseviert
40002:           DI[0] - DI[15]  - 16 Einzelbits für Kommunikation vom Slave zum Master
                 reseviert     
40003/40004:     float_reverse:   power_fwd
40005/40006:     float_reverse:   power_ref
40007/40008:     float_reverse:   power_drv
40009/40010:     float_reverse:   U1 (28V/48V)
40011/40012:     float_reverse:   I1 (ACS712,ACS715 o.ä.)
40013/40014:     float_reverse:   U2 (12V) 
40015/40016:     float_reverse:   temperature
40017/40018:     float_reverse:   SWR
40019/40020:     float_reverse:   res
MODBUS-Master: Kennung 0

Anzeige über Display LCD 2004


 */
// ---------------------------------------------- include libraries --------------------------------------------------
 #include <ModbusRtu.h>           //"Modbus-Master-Slave-for-Arduino-master" von smarmengol 
                                 //https://github.com/smarmengol/Modbus-Master-Slave-for-Arduino         
 #include <Wire.h>
 #include <LiquidCrystal_I2C.h>
 #include <stdlib.h>
 #include <DelayTimer.h>
 #include <Bounce2.h> 
// ---------------------------------------------- define constants ---------------------------------------------------
// Modbus 
  #define  modbus_baudrate 38400
  #define  modbus_serial Serial
  #define  modbus_txenable 2               // Pin 2 for enable TX in RS485
  
//Digital pins
  #define BUTTON_Right 3                  //Taster rechts 
  #define BUTTON_Left 4                   //Taster links
  #define BUTTON_Up 5                     //Taster hoch
  #define BUTTON_Down 6                   //Taster runter
  
  #define LED       13                    // interne LED      - Pin D13
 
// Debouncing   
  Bounce button_1 = Bounce();           // debouncing button1
  Bounce button_2 = Bounce();           // debouncing button2
  Bounce button_3 = Bounce();           // debouncing button3
  Bounce button_4 = Bounce();           // debouncing button4
  
// ---------------------------------------------- define variables ---------------------------------------------------
// Modbus 
  uint16_t au16data[25];                   // data array for modbus network sharing
  uint8_t u8state;                          // machine state
  uint8_t u8query;                          // pointer to message query
  unsigned long u32wait;
  uint8_t slave_nr = 1;                     //slave-nr: default = 1
  uint16_t pointer_offset;                  //offset for pointer to memory array (by slave-nr)
  
   unsigned long modbus_error_millis, modbus_error_delay_millis; 
   boolean state_modbus_error, state_modbus_error_old;
   
  
//shared memory for transfer float-variables via Modbus as float_reverse
  union memory {float floats[10]; word words[20];};

//values
 float FWD,REF,DRV,U1,I1,U2,T1,SWR,VALUE_8,Max_PWR;
 bool TX_ON;
 int bargraph;
 uint8_t page_nr = 1, page_nr_old = 0;
 uint8_t RX13 = 1, RX23 = 1, Cursor = 0;
  

// ---------------------------------------------- intialize instances -----------------------------------------------
  Modbus master(0,modbus_serial,modbus_txenable);     // this is master and RS485
  modbus_t telegram[2];                               // 2 telegrams

  /* Es gibt min. 2 Adressen welche ein Display verwenden kann: 0x3F oder 0x27.
     Setzt die PINs des I2C Moduls für die Kommunikation mit dem Display.
     Adresse: 0x27
     EN: 2
     RW; 1
     RS: 0
     D4: 4
     D5: 5
     D6: 6
     D7: 7
     BL: 3
     BLPOL: POSITIVE     */
  LiquidCrystal_I2C lcd(0x27, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE); //LCD-Display
  DelayTimer Band_Change_Timer;
  DelayTimer TimeOutLCD;
  boolean Input_state_change;  //Flag für Signalwechsel zum Aufwecken der Hintergrundbeleuchtung
 
  union memory transfer_values;


//################################################################################################################### 
void setup() 
{ 
//I/O-Pins
  pinMode(LED, OUTPUT);
  
// intialize Modbus-Master telegrams  
  pointer_offset = ((slave_nr - 1) * 20);
  
// telegram 0: read registers
  telegram[0].u8id = slave_nr; // slave address
  telegram[0].u8fct = 3; // function code (this one is registers read)
  telegram[0].u16RegAdd = 1; // start address in slave
  telegram[0].u16CoilsNo = 19; // number of elements (coils or registers) to read
  telegram[0].au16reg = au16data + 1; // pointer to a memory array in the Arduino
  
// telegram 1: write a single register
  telegram[1].u8id = slave_nr; // slave address
  telegram[1].u8fct = 6; // function code (this one is write a single register)
  telegram[1].u16RegAdd = 0; // start address in slave
  telegram[1].u16CoilsNo = 1; // number of elements (coils or registers) to read
  telegram[1].au16reg = au16data; // pointer to a memory array in the Arduino
  
// intialize the modbus-master
  modbus_serial.begin(modbus_baudrate);
  master.setTimeOut(2000); // if there is no answer in 2000 ms, roll over
  u32wait = millis() + 1000;
  u8state = u8query = 0; 
  master.start();
  
  digitalWrite(LED, HIGH);

  state_modbus_error = state_modbus_error_old = false;  
  modbus_error_delay_millis = millis() + 5000;           // delay of error-message   
//------------------------------------------------------------------------------------------------------------------
// intialize LCD-display    
 lcd.begin(20,4); // set up the LCD's number of columns and rows
 lcd.setCursor(0, 0);
 lcd.print("Telemetrie with"); 
 delay(500);
 lcd.setCursor(0, 1); 
 lcd.print("Arduino + Modbus/RTU"); 
 delay(500);
 lcd.setCursor(0, 2); 
 lcd.print("by DL1SUZ"); 
 delay(500);
 lcd.setCursor(11,2); 
 lcd.print("V 00.40"); 
 delay(2000);
 lcd.clear(); //Display loeschen 

  TimeOutLCD.setDelayTimer(600000); //600000 = 10min
  Input_state_change = false;
 

/* Vorbelegung für Subbandauswahl 13cm
   au16data[0]  bit 0: DO0
   au16data[0]  bit 1: DO1
   au16data[0]  bit 2: DO2
   au16dara[0]  bit 3: DO3
   bit = 0 entspricht DOx = HIGH
   bit = 1 entspricht DOx = LOW
   RX13=1: 2320MHz: DO0 = LOW ; DO1 = HIGH; [au16data0]= xxxx xxxx xxxx xx01
   RX13=2: 2400MHz: DO0 = HIGH; DO1 = HIGH; [au16data0]= xxxx xxxx xxxx xx00
   RX13=3: 2300MHz: DO0 = LOW ; DO1 = LOW ; [au16data0]= xxxx xxxx xxxx xx11
   RX13=4: 2304MHz: DO0 = HIGH; DO1 = LOW ; [au16data0]= xxxx xxxx xxxx xx10

 */
 au16data[0] = 0x0008;          //Vorbelegung der Variablen au16data[0] für RX 1296/2320MHz
                                //Bit3 gesetzt für Normalbetrieb Antennenrelais  
//setup debouncing button
  button_1.attach (BUTTON_Right ,INPUT_PULLUP);                         //PIN3
  button_1.interval (10);
  button_2.attach (BUTTON_Left ,INPUT_PULLUP);                          //PIN4
  button_2.interval (10);
  button_3.attach (BUTTON_Up ,INPUT_PULLUP);                            //PIN5
  button_3.interval (10);
  button_4.attach (BUTTON_Down ,INPUT_PULLUP);                          //PIN6
  button_4.interval (10);


// Bandweiterschaltung wenn Modbus-Teilnehmer nicht gefunden
   Band_Change_Timer.setDelayTimer(10000);    //30sec
  
  } 
// ---------------- End of setup ----------------------------------------
//################# Main loop ###################################################################################### 

void loop(){
//Debouncing buttons  
  button_1.update();                    // update bounce object 
  button_2.update();                    // update bounce object  
  button_3.update();                    // update bounce object   
  button_4.update();                    // update bounce object  
  lcd_dark();                           // Display nach 10min dunkelschalten
 
//main-program  
  modbus_master();

  value_mapping();
//Darstellung der Seiten 
 
  switch (page_nr){
  case 1:
  page1();                               // Anzeigen der Werte
  break;
  case 2:                               // Frequenzumschaltung 23cm
  page2();
  break;
  case 3:                               // Frequenzumschaltung 13cm
  page3();
  break;  
  }
//manuelle  Weiterschaltung der Slaves
  if (page_nr == 1){                           
  if (button_1.fell()){                 // fallende Flanke  Cursor rechts
  slave_nr = slave_nr + 1;
  }
  if (button_2.fell()){                 // fallende Flanke  Cursor links
  slave_nr = slave_nr - 1;
  }
  }
  
//automatische Suche nach aktivem Slave   
  if ((state_modbus_error == true) &&  Band_Change_Timer.isExpired()){
  slave_nr = slave_nr +1;
  Band_Change_Timer.setDelayTimer(10000); //10sec
  }
    
  if (slave_nr > 4) slave_nr = 1;
  if (slave_nr < 1) slave_nr = 4;
  
  telegram[0].u8id = slave_nr;
  telegram[1].u8id = slave_nr;
 
//SUB-Bandumschaltung für 23cm
  if ((slave_nr == 1)&&(page_nr == 1)){                   // Auf Seite 2 schalten       
  if (button_4.fell())  {                                 // fallende Flanke  Cursor runter
  page_nr = 2;
  Cursor = 0;
  }
  }
 
//SUB-Bandumschaltung für 13cm
  if ((slave_nr == 2)&&(page_nr == 1)){                   // Auf Seite 3 schalten       
  if (button_4.fell())  {                                 // fallende Flanke  Cursor runter
  page_nr = 3;
  Cursor = 0;
  }
  }

//Rückschaltung 23cm von Seite 2 auf Seite 1
  if ((slave_nr == 1)&&(page_nr == 2)&&(Cursor == 0)){   // Auf Seite 1 schalten       
  if (button_3.fell())  {                                // fallende Flanke  Cursor hoch
  page_nr = 1;
  Cursor = 0;
  }
  }

  //Rückschaltung 13cm von Seite 3 auf Seite 1
  if ((slave_nr == 2)&&(page_nr == 3)&&(Cursor == 0)){   // Auf Seite 1 schalten       
  if (button_3.fell())  {                                // fallende Flanke  Cursor hoch
  page_nr = 1;
  Cursor = 0;
  }
  }

//Zeilenweiterschaltung 23cm
  if ((slave_nr == 1)&&(page_nr == 2)&&(page_nr == page_nr_old)){                   // Cursor eine Zeile weiterschalten
  if (button_4.fell())  {                                 // fallende Flanke  Cursor runter
  Cursor = Cursor + 1;
  }
  }
  if (Cursor > 1) Cursor = 0;
  
//Zeilenweiterschaltung 13cm
 if ((slave_nr == 2)&&(page_nr == 3)&&(page_nr == page_nr_old)){                   // Cursor eine Zeile weiterschalten
  if (button_4.fell())  {                                 // fallende Flanke  Cursor runter
  Cursor = Cursor + 1;
  }
  }
  if (Cursor > 1) Cursor = 0;

//Umschaltung RX-Bereich 23cm
  if (slave_nr != 1) {goto m023;}                       
  if ((page_nr == 2)&&(Cursor == 1)){ 
  if (button_1.fell()){ 
  RX23 = RX23 + 1;
  }
  if (button_2.fell()){ 
  RX23 = RX23 - 1;
  }
  }
  if (RX23 > 2) RX23 = 1;
  if (RX23 < 1) RX23 = 2;

switch (RX23){
  case 1:
  bitWrite (au16data[0], 0, true);
  break;
  case 2:
  bitWrite (au16data[0], 0, false);
  break;
  }
m023: ;

//Umschaltung RX-Bereich 13cm
  if (slave_nr != 2) {goto m013;}                    
  if ((page_nr == 3)&&(Cursor == 1)){ 
  if (button_1.fell()){ 
  RX13 = RX13 + 1;
  }
  if (button_2.fell()){ 
  RX13 = RX13 - 1;
  }
  }
  if (RX13 > 4) RX13 = 1;
  if (RX13 < 1) RX13 = 4;

  switch (RX13){
  case 1:
  bitWrite (au16data[0], 0, true);
  bitWrite (au16data[0], 1, false);
  break;
  case 2:
  bitWrite (au16data[0], 0, false);
  bitWrite (au16data[0], 1, false);
  break;
  case 3:
  bitWrite (au16data[0], 0, true);
  bitWrite (au16data[0], 1, true);
  break;
  case 4:
  bitWrite (au16data[0], 0, false);
  bitWrite (au16data[0], 1, true);
  break;
  } 
m013: ;   
}
//################################################################################################################### 

//-------------------------------------------------------------------------------------------------------------------
// Modbus-Master nach einer Vorlage in [1] 
void modbus_master() 
{
 switch( u8state ) {
  case 0: 
    if (millis() > u32wait) {
    u8state++; // wait state
    digitalWrite(LED, LOW);
    }
    break;
  case 1: 
    master.query( telegram[u8query] ); // send query (only once)
    u8state++;
    u8query++;
  if (u8query > 1) u8query = 0;
    break;
  case 2:
    master.poll(); // check incoming messages
  if (master.getState() == COM_IDLE) {
     u8state = 0;
     u32wait = millis() + 250; 
     digitalWrite(LED, HIGH);
     }
    break;
 }
 // modbus-error-detection
   if (modbus_error_delay_millis > millis()) {goto m001;}     //startup delay for error-detection  
   if (master.getTimeOutState() == 1) {modbus_error_millis = millis();}
   if (modbus_error_millis + 3000 > millis()) {state_modbus_error = true; } //delaytime 3s
   else {state_modbus_error = false;} 
m001:  ;
}
//-------------------------------------------------------------------------------------------------------------------
void value_mapping()
{ 
  if (state_modbus_error == false) {  
    for (int i=0; i<=9; i++){
      transfer_values.words[(i*2)+1]= au16data[((i*2)) +2];  //Daten aus den Modbusregistern in den Transferbereich umkopieren
      transfer_values.words[(i*2)]= au16data[((i*2)+1) +2];  //dabei erfolgt ein Bytetausch 
    }  
    }
  else {                                                     //delete values if modbus-error
    for (int i=0; i<=9 ; i++){
    transfer_values.floats[i] = 0.0 ;
    }
  }
 
  FWD = transfer_values.floats[0];
  REF = transfer_values.floats[1];
  DRV = transfer_values.floats[2];
  U1 = transfer_values.floats[3];
  I1 = transfer_values.floats[4];
  U2 = transfer_values.floats[5];
  SWR = transfer_values.floats[6];
  T1 = transfer_values.floats[7];
  
  TX_ON = bitRead(au16data[1],0);     //PTT-Abfrage
  
}            
//-------------------------------------------------------------------------------------------------------------------
void page1()
{
//Seite löschen bei erstmaligem Aufruf und feste Bestandteile schreiben
   if (page_nr != page_nr_old){
    lcd.clear();
//Zeile 2 zusammenstellen
  lcd.setCursor(0, 1);
  lcd.print ("U1: ");
  lcd.setCursor(8, 1);
  lcd.print ("V");
  lcd.setCursor(10, 1);
  lcd.print ("FWD:");
  lcd.setCursor(18, 1);
  lcd.print ("W");
//Zeile 3 zusammenstellen 
  lcd.setCursor(0, 2); 
  lcd.print ("I1:"); 
  lcd.setCursor(8, 2);
  lcd.print ("A");  
  lcd.setCursor(10, 2);  
  lcd.print ("REF:"); 
  lcd.setCursor(18, 2);
  lcd.print ("W");   
   }
   page_nr_old = page_nr;
   
//Variablendefinitionen
   char buffer_1[5];   
   char buffer_2[5];  
   char buffer_3[5];
   char buffer_4[5]; 
   char buffer_5[5];
   char buffer_6[5]; 
   char buffer_7[5]; 
   char buffer_8[5]; 
   String state;

//Ausgabestrings zusammenstellen  
//Floatwerte in String wandeln
  dtostrf(U1, 4, 1, buffer_1);                            //Wandlung Float > String Laenge 4 Zeichen 1 Nachkommastelle
  if (I1 < 100.0)  dtostrf(I1, 4, 1, buffer_2); 
    else dtostrf(I1, 4, 0, buffer_2); 
  dtostrf(U2, 4, 1, buffer_3);   
  dtostrf(DRV, 4, 1, buffer_4);  
  dtostrf(FWD, 4, 0, buffer_5); 
  if (REF < 10.0)  dtostrf(REF, 4, 1, buffer_6); 
    else  dtostrf(REF, 4, 0, buffer_6); 
  if (T1 < -99.9) dtostrf(T1, 4, 0, buffer_7);
    else dtostrf(T1, 4, 1, buffer_7); 
  dtostrf(SWR, 4, 1, buffer_8);
  
//state zusammenstellen  
  if (state_modbus_error == true) state = "Modbus-Error";
  else {
  if (TX_ON == true) state = "TX";
  else state = "RX"; 
  }

// Displayausgabe
  if (state_modbus_error != state_modbus_error_old) {
  lcd.clear();    //clean for error-screen
  page_nr_old = 0;
  } 
  state_modbus_error_old = state_modbus_error;

//Zeile 1 zusammenstellen
  lcd.setCursor(0, 0);
  switch (slave_nr){    //Ausgabe der Bandbezeichnung in Abhängigkeit der Slave-Nr
  case 1:
  lcd.print ("23cm:");
  Max_PWR = 400;
  break;
  case 2:
  lcd.print ("13cm:");
  Max_PWR = 200;
  break;
  case 3:
  lcd.print (" 9cm:");
  Max_PWR = 150;
  break; 
  case 4:
  lcd.print (" 6cm:");
   Max_PWR = 50;
  break;   
  }
  
  lcd.setCursor(6, 0); 
  lcd.print (state); 

  if (state_modbus_error == false) {
  lcd.setCursor(10, 0);  
  lcd.print ("Tmp:"); 
  lcd.setCursor(14, 0);
  lcd.print (buffer_7); 
  lcd.setCursor(18, 0);
  lcd.print (char (223));
  lcd.print ("C");
  }
  
//Zeile 2 zusammenstellen 
  lcd.setCursor(4, 1);
  lcd.print (buffer_1); 
  lcd.setCursor(14, 1);
  lcd.print (buffer_5);

//Zeile 3 zusammenstellen 
  lcd.setCursor(4, 2);
  lcd.print (buffer_2); 
  lcd.setCursor(14, 2);
  lcd.print (buffer_6); 
   
//Zeile 4 zusammenstellen
  if (TX_ON == false){
  lcd.setCursor(0, 3);
  lcd.print ("U2: ");
  lcd.setCursor(4, 3);
  lcd.print (buffer_3); 
  lcd.setCursor(8, 3);
  lcd.print ("V");
  }
  else  {
  lcd.setCursor(0, 3);
  lcd.print ("SWR:");
  lcd.setCursor(4, 3);
  lcd.print (buffer_8); 
  lcd.setCursor(8, 3);
  lcd.print ("  ");  
 
  }
//bargraph   
  bargraph = map(FWD, 0, Max_PWR, 0, 10);
  lcd.setCursor(10, 3);
  for (int z=0; z<10; ++z) {
  if (z < bargraph) { 
       lcd.print(char (0xFF));  //volles Feld
      }
      else {
       lcd.print(char (0x20));  //Leerzeichen
      }
    }   
}

//-------------------------------------------------------------------------------------------------------------------
void page2(){
  if (page_nr != page_nr_old){
   lcd.clear();
//Zeile 1 zusammenstellen
  lcd.setCursor(0, 0);
  lcd.print ("23cm subband");
 
//Zeile 2 zusammenstellen 
  lcd.setCursor(0, 1);
  lcd.print ("RX: ");
  lcd.setCursor(9, 1);
  lcd.print ("MHz");
 }  
  page_nr_old = page_nr;
  
//Zeile 2
  lcd.setCursor(5, 1);
  switch (RX23){
  case 1:
  lcd.print ("1296");  
  break;
  case 2:
  lcd.print ("1268");  
  break;
  }
    
//Cursor darstellen 
  lcd.setCursor(18,0);
  if  (Cursor == 0) lcd.print("X");
  else lcd.print(" "); 
    
  lcd.setCursor(18,1);
  if (Cursor == 1)lcd.print("X");
  else lcd.print(" "); 
}
//-------------------------------------------------------------------------------------------------------------------
void page3(){
   if (page_nr != page_nr_old){
   lcd.clear();
//Zeile 1 zusammenstellen
  lcd.setCursor(0, 0);
  lcd.print ("13cm subband");
 
//Zeile 2 zusammenstellen 
  lcd.setCursor(0, 1);
  lcd.print ("RX: ");
  lcd.setCursor(9, 1);
  lcd.print ("MHz");
 }  
  page_nr_old = page_nr;
  
//Zeile 2
  lcd.setCursor(5, 1);
  switch (RX13){
  case 1:
  lcd.print ("2320");  
  break;
  case 2:
  lcd.print ("2400");  
  break;
  case 3:
  lcd.print ("2300");  
  break;
  case 4:
  lcd.print ("2304");  
  break;  
  }
  
//Cursor darstellen 
  lcd.setCursor(18,0);
  if  (Cursor == 0) lcd.print("X");
  else lcd.print(" "); 
    
  lcd.setCursor(18,1);
  if (Cursor == 1)lcd.print("X");
  else lcd.print(" "); 
}
//-------------------------------------------------------------------------------------------------------------------
  void lcd_dark(){
  if (button_1.fell()|| button_2.fell() || button_3.fell() || button_4.fell() || (TX_ON == true)) {
  Input_state_change = true;
  }
  if (Input_state_change == true){
  TimeOutLCD.setDelayTimer(600000);     //Zeit in ms: 600000ms = 10min
  } 
  if (TimeOutLCD.isExpired()){
  lcd.noBacklight();
  } 
  else {
  lcd.backlight(); 
  }
  Input_state_change = false;       //Flag für Erkennung des Signalwechsels Reset    
  } 
//-------------------------------------------------------------------------------------------------------------------
